/* ===========================================================================
 * $RCSfile: ClassTree.java,v $
 * ===========================================================================
 *
 * RetroGuard -- an obfuscation package for Java classfiles.
 *
 * Copyright (c) 1999 Mark Welsh (markw@retrologic.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * The author may be contacted at markw@retrologic.com 
 *
 *
 * $Date: 1999/05/23 20:50:55 $
 * $Revision: 1.4 $
 */
package COM.rl.obf;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import COM.rl.util.*;
import COM.rl.obf.classfile.*;

/**
 * Tree structure of package levels, classes, methods and fields used for obfuscation.
 *
 * @author      Mark Welsh
 */
public class ClassTree implements NameMapper
{
    // Constants -------------------------------------------------------------
    public static final char PACKAGE_LEVEL = '/';
    public static final char CLASS_LEVEL = '$';
    public static final char METHOD_FIELD_LEVEL = '/';
    private static final String LOG_PRE_UNOBFUSCATED =
        "# Names reserved from obfuscation:";
    private static final String LOG_PRE_OBFUSCATED =
        "# Obfuscated name mappings (some of these may be unchanged due to polymorphism constraints):";
    private static final String LOG_FREQUENCY_TABLE =
        "# Obfuscated name overloading frequency:";


    // Fields ----------------------------------------------------------------
    private Vector retainAttrs = new Vector();  // List of attributes to retain
    private Pk root = null;   // Root package in database (Java default package)

    // Class methods ---------------------------------------------------------
    /** Return a fully qualified name broken into package/class segments. */
    public static Enumeration getNameEnum(String name) throws Exception
    {
        Vector vec = new Vector();
        while (!name.equals(""))
        {
            int posP = name.indexOf(PACKAGE_LEVEL);
            int posC = name.indexOf(CLASS_LEVEL);
            Cons cons = null;
            if (posP == -1 && posC == -1)
            {
                cons = new Cons(new Character(CLASS_LEVEL), name);
                name = "";
            }
            if (posP == -1 && posC != -1)
            {
                cons = new Cons(new Character(CLASS_LEVEL), name.substring(0, posC));
                name = name.substring(posC + 1, name.length());
            }
            if (posP != -1 && posC == -1)
            {
                cons = new Cons(new Character(PACKAGE_LEVEL), name.substring(0, posP));
                name = name.substring(posP + 1, name.length());
            }
            if (posP != -1 && posC != -1)
            {
                if (posP < posC)
                {
                    cons = new Cons(new Character(PACKAGE_LEVEL), name.substring(0, posP));
                    name = name.substring(posP + 1, name.length());
                }
                else
                {
                    throw new IOException("Invalid fully qualified name");
                }
            }
            if (((String)cons.cdr).equals(""))
            {
                throw new IOException("Invalid fully qualified name");
            }
            vec.addElement(cons);
        }
        return vec.elements();
    }


    // Instance Methods ------------------------------------------------------
    /** Ctor. */
    public ClassTree()
    {
        root = Pk.createRoot(this);
    }

    /** Return the root node. */
    public Pk getRoot() {return root;}

    /** Update the path of the passed filename, if that path corresponds to a package. */
    public String getOutName(String inName) 
    {
        try
        {
            TreeItem ti = root;
            StringBuffer sb = new StringBuffer();
            for (Enumeration nameEnum = getNameEnum(inName); nameEnum.hasMoreElements(); )
            {
                Cons nameSegment = (Cons)nameEnum.nextElement();
                char tag = ((Character)nameSegment.car).charValue();
                String name = (String)nameSegment.cdr;
                switch (tag)
                {
                case PACKAGE_LEVEL:
                    if (ti != null)
                    {
                        ti = ((Pk)ti).getPackage(name);
                        if (ti != null)
                        {
                            sb.append(ti.getOutName());
                        }
                        else
                        {
                            sb.append(name); 
                        }
                    }
                    else
                    {
                        sb.append(name);
                    }
                    sb.append(PACKAGE_LEVEL);
                    break;
                    
                case CLASS_LEVEL:
                    sb.append(name);
                    return sb.toString();
                    
                default:
                    throw new Exception("Internal error: illegal package/class name tag");
                }
            }
        }
        catch (Exception e)
        {
            // Just drop through and return the original name
        }
        return inName;
    }

    /** Add a classfile's package, class, method and field entries to database. */
    public void addClassFile(ClassFile cf) throws Exception
    {
        // Add the fully qualified class name
        TreeItem ti = root;
        char parentTag = PACKAGE_LEVEL;
        for (Enumeration nameEnum = getNameEnum(cf.getName()); nameEnum.hasMoreElements(); )
        {
            Cons nameSegment = (Cons)nameEnum.nextElement();
            char tag = ((Character)nameSegment.car).charValue();
            String name = (String)nameSegment.cdr;
            switch (tag)
            {
            case PACKAGE_LEVEL:
                ti = ((Pk)ti).addPackage(name);
                break;

            case CLASS_LEVEL:
                // If this is an inner class, just add placeholder classes up the tree
                if (nameEnum.hasMoreElements())
                {
                    ti = ((PkCl)ti).addPlaceholderClass(name);
                }
                else
                {
                    ti = ((PkCl)ti).addClass(name, cf.getSuper(), cf.getInterfaces());
                }
                break;

            default:
                throw new Exception("Internal error: illegal package/class name tag");
            }
            parentTag = tag;
        }

        // We must have a class before adding methods and fields
        if (ti instanceof Cl)
        {
            Cl cl = (Cl)ti;

            // Add the class's methods to the database
            for (Enumeration enum = cf.getMethodEnum(); enum.hasMoreElements(); )
            {
                Cons cons = (Cons)enum.nextElement();
                cl.addMethod(((Boolean)((Cons)cons.car).car).booleanValue(), (String)((Cons)cons.car).cdr, (String)((Cons)cons.cdr).car, ((Integer)((Cons)cons.cdr).cdr).intValue());
            }

            // Add the class's fields to the database
            for (Enumeration enum = cf.getFieldEnum(); enum.hasMoreElements(); )
            {
                Cons cons = (Cons)enum.nextElement();
                cl.addField(((Boolean)((Cons)cons.car).car).booleanValue(), (String)((Cons)cons.car).cdr, (String)((Cons)cons.cdr).car, ((Integer)((Cons)cons.cdr).cdr).intValue());
            }
        }
        else
        {
            throw new Exception("Inconsistent class file.");
        }
    }

    /** Mark an attribute type for retention. */
    public void retainAttribute(String name) throws Exception
    {
        retainAttrs.addElement(name);
    }

    /** Mark a class/interface type (and possibly methods and fields defined in class) for retention. */
    public void retainClass(String name, boolean retainToPublic,
                                         boolean retainToProtected,
                                         boolean retainFieldsOnly,
                                         boolean retainMethodsOnly) throws Exception
    {
        // Mark the class (or classes, if this is a wildcarded specifier)
        for (Enumeration clEnum = getClEnum(name); clEnum.hasMoreElements(); )
        {
            Cl classItem = (Cl)clEnum.nextElement();
            retainHierarchy(classItem);
            if (retainToPublic || retainToProtected)
            {
                // Retain methods if requested
                if (!retainFieldsOnly)
                {
                    for (Enumeration enum = classItem.getMethodEnum(); enum.hasMoreElements(); )
                    {
                        Md md = (Md)enum.nextElement();
                        if ((retainToPublic && Modifier.isPublic(md.getModifiers())) ||
                            (retainToProtected && !Modifier.isPrivate(md.getModifiers())))
                        {
                            md.setOutName(md.getInName());
                            md.setFromScript();
                        }
                    }
                }

                // Retain fields if requested
                if (!retainMethodsOnly)
                {
                    for (Enumeration enum = classItem.getFieldEnum(); enum.hasMoreElements(); )
                    {
                        Fd fd = (Fd)enum.nextElement();
                        if ((retainToPublic && Modifier.isPublic(fd.getModifiers())) ||
                            (retainToProtected && !Modifier.isPrivate(fd.getModifiers())))
                        {
                            fd.setOutName(fd.getInName());
                            fd.setFromScript();
                        }
                    }
                }
            }
        }
    }

    /** Mark a method type for retention. */
    public void retainMethod(String name, String descriptor) throws Exception
    {
        Md md = getMd(name, descriptor);
        md.setOutName(md.getInName());
        md.setFromScript();
    }

    /** Mark a field type for retention. */
    public void retainField(String name) throws Exception
    {
        Fd fd = getFd(name);
        fd.setOutName(fd.getInName());
        fd.setFromScript();
    }

    /** Mark a package for retention, and specify its new name. */
    public void retainPackageMap(String name, String obfName) throws Exception
    {
        retainItemMap(getPk(name), obfName);
    }

    /** Mark a class/interface type for retention, and specify its new name. */
    public void retainClassMap(String name, String obfName) throws Exception
    {
        retainItemMap(getCl(name), obfName);
    }

    /** Mark a method type for retention, and specify its new name. */
    public void retainMethodMap(String name, String descriptor, 
                                String obfName) throws Exception
    {
        retainItemMap(getMd(name, descriptor), obfName);
    }

    /** Mark a field type for retention, and specify its new name. */
    public void retainFieldMap(String name, String obfName) throws Exception
    {
        retainItemMap(getFd(name), obfName);
    }

    // Mark an item for retention, and specify its new name.
    private void retainItemMap(TreeItem item, String obfName) throws Exception
    {
        if (!item.isFixed()) 
        {
            item.setOutName(obfName);
            item.setFromScriptMap();
        }
    }

    /** Traverse the class tree, generating obfuscated names within each namespace. */
    public void generateNames() throws Exception
    {
        walkTree(new TreeAction() {
            public void packageAction(Pk pk) throws Exception {pk.generateNames();}
            public void classAction(Cl cl) throws Exception {cl.generateNames();}
        });
    }

    /** Resolve the polymorphic dependencies of each class. */
    public void resolveClasses() throws Exception
    {
        walkTree(new TreeAction() {
            public void classAction(Cl cl) throws Exception {cl.resetResolve();}
        });
        walkTree(new TreeAction() {
            public void classAction(Cl cl) throws Exception {cl.setupNameListDowns();}
        });
        Cl.nameSpace = 0;
        walkTree(new TreeAction() {
            public void classAction(Cl cl) throws Exception {cl.resolveOptimally();}
        });
    }

    /** Return a list of attributes marked to keep. */
    public String[] getAttrsToKeep() throws Exception
    {
        String[] attrs = new String[retainAttrs.size()];
        for (int i = 0; i < attrs.length; i++)
        {
            attrs[i] = (String)retainAttrs.elementAt(i);
        }
        return attrs;
    }

    /** Get classes in tree from the fully qualified name (can be wildcarded at the end). */
    public Enumeration getClEnum(String fullName) throws Exception
    {
        Enumeration clEnum = null;

        // Wildcarded? If so, return list of all classes (including inner classes) in package
        if (fullName.charAt(fullName.length() - 1) == '*')
        {
            String pkName = "";
            if (fullName.length() > 2)
            {
                pkName = fullName.substring(0, fullName.length() - 2);
            }
            Pk pk = getPk(pkName);
            if (pk != null)
            {
                clEnum = pk.getAllClassEnum();
            }
        }
        else
        {
            // Single class
            Vector classes = new Vector();
            Cl cl = getCl(fullName);
            if (cl != null)
            {
                classes.addElement(cl);
            }
            clEnum = classes.elements();
        }
        if (clEnum == null)
        {
            clEnum = new Vector().elements();
        }
        return clEnum;
    }

    /** Get class in tree from the fully qualified name, returning null if name not found. */
    public Cl getCl(String fullName) throws Exception
    {
        TreeItem ti = root;
        for (Enumeration nameEnum = getNameEnum(fullName); nameEnum.hasMoreElements(); )
        {
            Cons nameSegment = (Cons)nameEnum.nextElement();
            char tag = ((Character)nameSegment.car).charValue();
            String name = (String)nameSegment.cdr;
            switch (tag)
            {
            case PACKAGE_LEVEL:
                ti = ((Pk)ti).getPackage(name);
                break;

            case CLASS_LEVEL:
                ti = ((PkCl)ti).getClass(name);
                break;

            default:
                throw new Exception("Internal error: illegal package/class name tag");
            }

            // If the name is not in the database, return null
            if (ti == null)
            {
                return null;
            }
        }

        // It is an error if we do not end up with a class or interface
        if (!(ti instanceof Cl))
        {
            throw new Exception("Inconsistent class or interface name.");
        }
        return (Cl)ti;
    }

    /** Get package in tree from the fully qualified name, returning null if name not found. */
    public Pk getPk(String fullName) throws Exception
    {
        TreeItem ti = root;
        for (Enumeration nameEnum = getNameEnum(fullName); nameEnum.hasMoreElements(); )
        {
            Cons nameSegment = (Cons)nameEnum.nextElement();
            String name = (String)nameSegment.cdr;
            ti = ((Pk)ti).getPackage(name);

            // If the name is not in the database, return null
            if (ti == null)
            {
                return null;
            }
            // It is an error if we do not end up with a package
            if (!(ti instanceof Pk))
            {
                throw new Exception("Inconsistent package.");
            }
        }
        return (Pk)ti;
    }

    /** Get method in tree from the fully qualified name. */
    public Md getMd(String fullName, String descriptor) throws Exception
    {
        // Split into class and method names
        int pos = fullName.lastIndexOf(METHOD_FIELD_LEVEL);
        Cl cl = getCl(fullName.substring(0, pos));
        return cl.getMethod(fullName.substring(pos + 1), descriptor);
    }

    /** Get field in tree from the fully qualified name. */
    public Fd getFd(String fullName) throws Exception
    {
        // Split into class and field names
        int pos = fullName.lastIndexOf(METHOD_FIELD_LEVEL);
        Cl cl = getCl(fullName.substring(0, pos));
        return cl.getField(fullName.substring(pos + 1));
    }

    /** Mapping for fully qualified class name.
     *  @see NameMapper#mapClass */
    public String mapClass(String className) throws Exception
    {
        // Check for array -- requires special handling
        if (className.length() > 0 && className.charAt(0) == '[') {
            StringBuffer newName = new StringBuffer();
            int i = 0;
            while (i < className.length()) {
                char ch = className.charAt(i++);
                switch (ch) {
                case '[':
                case ';':
                    newName.append(ch);
                    break;

                case 'L':
                    newName.append(ch);
                    int pos = className.indexOf(';', i);
                    if (pos < 0) {
                        throw new Exception("Invalid class name encountered: " + className);
                    }
                    newName.append(mapClass(className.substring(i, pos)));
                    i = pos;
                    break;
                    
                default:
                    return className;
                }
            }
            return newName.toString();
        } else {
            Cl cl = getCl(className);
            return cl != null ? cl.getFullOutName() : className;
        }
    }

    /** Mapping for method name, of fully qualified class.
     *  @see NameMapper#mapMethod */
    public String mapMethod(String className, String methodName, String descriptor) throws Exception
    {
        Cl cl = getCl(className);
        return cl != null && cl.getMethod(methodName, descriptor) != null ?
                    cl.getMethod(methodName, descriptor).getOutName() :
                    methodName;
    }

    /** Mapping for field name, of fully qualified class.
     *  @see NameMapper#mapField */
    public String mapField(String className, String fieldName) throws Exception
    {
        Cl cl = getCl(className);
        return cl != null && cl.getField(fieldName) != null ?
                    cl.getField(fieldName).getOutName() :
                    fieldName;
    }

    /** Mapping for descriptor of field or method.
     *  @see NameMapper#mapDescriptor */
    public String mapDescriptor(String descriptor) throws Exception
    {
        // Pass everything through unchanged, except for the String between
        // 'L' and ';' -- this is passed through mapClass(String)
        StringBuffer newDesc = new StringBuffer();
        int i = 0;
        while (i < descriptor.length())
        {
            char ch = descriptor.charAt(i++);
            switch (ch)
            {
            case '[':
            case 'B':
            case 'C':
            case 'D':
            case 'F':
            case 'I':
            case 'J':
            case 'S':
            case 'Z':
            case 'V':
            case '(':
            case ')':
            case ';':
                newDesc.append(ch);
                break;

            case 'L':
                newDesc.append(ch);
                int pos = descriptor.indexOf(';', i);
                if (pos < 0)
                {
                    throw new Exception("Invalid descriptor string encountered.");
                }
                newDesc.append(mapClass(descriptor.substring(i, pos)));
                i = pos;
                break;

            default:
                throw new Exception("Invalid descriptor string encountered.");
            }
        }
        return newDesc.toString();
    }

    /** Dump the content of the class tree to the specified file (used for logging). */
    public void dump(final PrintWriter log) throws Exception
    {
        log.println("#");
        log.println(LOG_FREQUENCY_TABLE);
        dumpFrequencyCount(log);
        log.println("#");
        log.println(LOG_PRE_UNOBFUSCATED);
        log.println("#");
        walkTree(new TreeAction() {
            public void classAction(Cl cl) {
                if (cl.isFromScript()) {
                    log.println(".class " + cl.getFullInName());
                }
            }
            public void methodAction(Md md) {
                if (md.isFromScript()) {
                    log.println(".method " + md.getFullInName() + " " + md.getDescriptor());
                }
            }
            public void fieldAction(Fd fd) {
                if (fd.isFromScript()) {
                    log.println(".field " + fd.getFullInName() + " " + fd.getDescriptor());
                }
            }
            public void packageAction(Pk pk) {
                // No action
            }
        });
        log.println("#");
        log.println("#");
        log.println(LOG_PRE_OBFUSCATED);
        log.println("#");
        walkTree(new TreeAction() {
            public void classAction(Cl cl) {
                if (!cl.isFromScript()) {
                    log.println(".class_map " + cl.getFullInName() + " " + cl.getOutName());
                }
            }
            public void methodAction(Md md) {
                if (!md.isFromScript()) {
                    log.println(".method_map " + md.getFullInName() + " " + md.getDescriptor() + " " + md.getOutName());
                }
            }
            public void fieldAction(Fd fd) {
                if (!fd.isFromScript()) {
                    log.println(".field_map " + fd.getFullInName() + " " + fd.getOutName());
                }
            }
            public void packageAction(Pk pk) {
                if (!pk.isFromScript() && pk.getFullInName().length() > 0) {
                    log.println(".package_map " + pk.getFullInName() + " " + pk.getOutName());
                }
            }
        });
    }
    private void dumpFrequencyCount(PrintWriter log)
    {
        // Compute total use count
        int totalUseCount = 0;
        for (Enumeration useCountEnum = NameMaker.getUseCounts(); useCountEnum.hasMoreElements(); )
        {
            totalUseCount += ((Integer)useCountEnum.nextElement()).intValue() + 1;
        }

        // Log the individual use counts for names
        if (totalUseCount != 0)
        {
            int sumPercent = 0;
            int sumUseCount = 0;
            Vector sort = new Vector();
            for (Enumeration nameEnum = NameMaker.getNames(), useCountEnum = NameMaker.getUseCounts(); nameEnum.hasMoreElements(); )
            {
                String name = (String)nameEnum.nextElement();
                int useCount = ((Integer)useCountEnum.nextElement()).intValue() + 1;
                int percent = 100 * useCount / totalUseCount;
                if (percent != 0)
                {
                    sort.addElement(new SortElement(name, useCount));
                    sumUseCount += useCount;
                    sumPercent += percent;
                }
            }
            while (sort.size() != 0)
            {
                SortElement se = null;
                int largestUseCount = 0;
                for (Enumeration enum = sort.elements(); enum.hasMoreElements(); )
                {
                    SortElement thisSe = (SortElement)enum.nextElement();
                    if (thisSe.useCount >= largestUseCount)
                    {
                        se = thisSe;
                        largestUseCount = se.useCount;
                    }
                }
                sort.removeElement(se);
                log.println("#  '" + se.name + "'   \tused " + se.useCount + " times\t(" + Integer.toString(100 * se.useCount / totalUseCount) + "%)");
            }

            // Log the remainder percentage
            log.println("#  Other names (each used in <1% of mappings) used a total of " +
                        Integer.toString(totalUseCount - sumUseCount) +
                        " times (" + Integer.toString(100 - sumPercent) + "%)");
            log.println("#");
        }
    }
    class SortElement
    {
        int useCount;
        String name;
        SortElement(String name, int useCount) {this.useCount = useCount; this.name = name;}
    }

    // Private Methods -------------------------------------------------------
    // Mark TreeItem and all parents for retention.
    private void retainHierarchy(TreeItem ti) throws Exception
    {
        if (!ti.isFixed()) 
        {
            ti.setOutName(ti.getInName());
            ti.setFromScript();
        }
        if (ti.parent != null)
        {
            retainHierarchy(ti.parent);
        }
    }

    /** Walk the whole tree taking action once only on each package level, class, method and field. */
    public void walkTree(TreeAction ta) throws Exception
    {
        walkTree(ta, root);
    }

    // Walk the tree which has TreeItem as its root taking action once only on each
    // package level, class, method and field.
    private void walkTree(TreeAction ta, TreeItem ti) throws Exception
    {
        if (ti instanceof Pk)
        {
            Enumeration packageEnum = ((Pk)ti).getPackageEnum();
            ta.packageAction((Pk)ti);
            while (packageEnum.hasMoreElements())
            {
                walkTree(ta, (TreeItem)packageEnum.nextElement());
            }
        }
        if (ti instanceof PkCl)
        {
            Enumeration classEnum = ((PkCl)ti).getClassEnum();
            while (classEnum.hasMoreElements())
            {
                walkTree(ta, (TreeItem)classEnum.nextElement());
            }
        }
        if (ti instanceof Cl)
        {
            Enumeration fieldEnum = ((Cl)ti).getFieldEnum();
            Enumeration methodEnum = ((Cl)ti).getMethodEnum();
            ta.classAction((Cl)ti);
            while (fieldEnum.hasMoreElements())
            {
                ta.fieldAction((Fd)fieldEnum.nextElement());
            }
            while (methodEnum.hasMoreElements())
            {
                ta.methodAction((Md)methodEnum.nextElement());
            }
        }
    }
}


