/* ===========================================================================
 * $RCSfile: RgsEnum.java,v $
 * ===========================================================================
 *
 * RetroGuard -- an obfuscation package for Java classfiles.
 *
 * Copyright (c) 1999 Mark Welsh (markw@retrologic.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * The author may be contacted at markw@retrologic.com 
 *
 *
 * $Date: 1998/10/01 14:14:13 $
 * $Revision: 1.3 $
 */
package COM.rl.obf;

import java.io.*;
import java.util.*;
import COM.rl.util.*;
import COM.rl.obf.classfile.*;

/**
 * Parser for RGS script files which provides an enumeration of option entries.
 *
 * @author      Mark Welsh
 */
public class RgsEnum
{
    // Constants -------------------------------------------------------------
    public static final String DIRECTIVE_ATTR = ".attribute";
    public static final String DIRECTIVE_CLASS = ".class";
    public static final String DIRECTIVE_METHOD = ".method";
    public static final String DIRECTIVE_FIELD = ".field";
    public static final String DIRECTIVE_PACKAGE_MAP = ".package_map";
    public static final String DIRECTIVE_CLASS_MAP = ".class_map";
    public static final String DIRECTIVE_METHOD_MAP = ".method_map";
    public static final String DIRECTIVE_FIELD_MAP = ".field_map";
    public static final String OPTION_PUBLIC = "public";
    public static final String OPTION_PROTECTED = "protected";
    public static final String OPTION_METHOD = "method";
    public static final String OPTION_FIELD = "field";


    // Fields ----------------------------------------------------------------
    private StreamTokenizer tk;
    private RgsEntry next;
    private Exception nextException;


    // Class Methods ---------------------------------------------------------


    // Instance Methods ------------------------------------------------------
    /** Ctor. */
    public RgsEnum(InputStream rgs)
    {
        tk = new StreamTokenizer(
             new BufferedReader(
             new InputStreamReader(rgs)));
        tk.resetSyntax();
        tk.whitespaceChars(0x00, 0x20);
        tk.wordChars('*', '*');
        tk.wordChars('.', '.');
        tk.wordChars(';', ';');
        tk.wordChars('_', '_');
        tk.wordChars('[', '[');
        tk.wordChars('(', ')');
        tk.wordChars('$', '$');
        tk.wordChars('/', '9');
        tk.wordChars('A', 'Z');
        tk.wordChars('a', 'z');
        tk.commentChar('#');
        tk.eolIsSignificant(true);
        readNext();
    }

    /** Are there more script entries? */
    public boolean hasMoreEntries() throws Exception
    {
        if (nextException != null)
        {
            throw nextException;
        }
        return next != null;
    }

    /** Return next script entry. */
    public RgsEntry nextEntry() throws Exception
    {
        RgsEntry thisOne = next;
        Exception thisException = nextException;
        readNext();
        if (thisException != null)
        {
            throw thisException;
        }
        return thisOne;
    }

    // Read the next entry, returning true if one is available.
    private void readNext() {
        // Reset the 'next error' state
        nextException = null;
        RgsEntry entry = null;
        int ttype;
        try {
            int directive = -1;
            String name = null;
            String descriptor = null;
            while ((ttype = tk.nextToken()) != StreamTokenizer.TT_EOF) {
                if (ttype == StreamTokenizer.TT_WORD) {
                    if (directive == -1) {
                        if (tk.sval.equals(DIRECTIVE_ATTR)) {
                            directive = RgsEntry.TYPE_ATTR;
                        } else if (tk.sval.equals(DIRECTIVE_CLASS)) {
                            directive = RgsEntry.TYPE_CLASS;
                        } else if (tk.sval.equals(DIRECTIVE_METHOD)) {
                            directive = RgsEntry.TYPE_METHOD;
                        } else if (tk.sval.equals(DIRECTIVE_FIELD)) {
                            directive = RgsEntry.TYPE_FIELD;
                        } else if (tk.sval.equals(DIRECTIVE_PACKAGE_MAP)) {
                            directive = RgsEntry.TYPE_PACKAGE_MAP;
                        } else if (tk.sval.equals(DIRECTIVE_CLASS_MAP)) {
                            directive = RgsEntry.TYPE_CLASS_MAP;
                        } else if (tk.sval.equals(DIRECTIVE_METHOD_MAP)) {
                            directive = RgsEntry.TYPE_METHOD_MAP;
                        } else if (tk.sval.equals(DIRECTIVE_FIELD_MAP)) {
                            directive = RgsEntry.TYPE_FIELD_MAP;
                        } else {
                            throw new Exception();
                        }
                    } else if (entry == null) {
                        if (directive == RgsEntry.TYPE_METHOD) {
                            if (name == null) {
                                name = tk.sval;
                                checkMethodOrFieldSpec(name);
                            } else {
                                descriptor = tk.sval;
                                checkMethodDescriptor(descriptor);
                                entry = new RgsEntry(directive, name, descriptor);
                            }
                        } else if (directive == RgsEntry.TYPE_FIELD) {
                            if (name == null) {
                                name = tk.sval;
                                checkMethodOrFieldSpec(name);
                            } else {
                                descriptor = tk.sval;
                                checkJavaType(descriptor);
                                entry = new RgsEntry(directive, name, descriptor);
                            }
                        } else if (directive == RgsEntry.TYPE_ATTR) {
                            if (!Tools.isInArray(tk.sval, ClassConstants.KNOWN_ATTRS)) {
                                throw new Exception();
                            }
                            entry = new RgsEntry(directive, tk.sval);
                        } else if (directive == RgsEntry.TYPE_CLASS) {
                            checkClassWCSpec(tk.sval);
                            entry = new RgsEntry(directive, tk.sval);
                        } else if (directive == RgsEntry.TYPE_PACKAGE_MAP) {
                            if (name == null) {
                                name = tk.sval;
                                checkClassSpec(name);
                            } else {
                                String obfName = tk.sval;
                                checkJavaIdentifier(obfName);
                                entry = new RgsEntry(directive, name);
                                entry.obfName = obfName;
                            }
                        } else if (directive == RgsEntry.TYPE_CLASS_MAP) {
                            if (name == null) {
                                name = tk.sval;
                                checkClassSpec(name);
                            } else {
                                String obfName = tk.sval;
                                checkJavaInnerIdentifier(obfName);
                                entry = new RgsEntry(directive, name);
                                entry.obfName = obfName;
                            }
                        } else if (directive == RgsEntry.TYPE_METHOD_MAP) {
                            if (name == null) {
                                name = tk.sval;
                                checkMethodOrFieldSpec(name);
                            } else if (descriptor == null) {
                                descriptor = tk.sval;
                                checkMethodDescriptor(descriptor);
                            } else {
                                String obfName = tk.sval;
                                checkJavaIdentifier(obfName);
                                entry = new RgsEntry(directive, name, descriptor);
                                entry.obfName = obfName;
                            }
                        } else if (directive == RgsEntry.TYPE_FIELD_MAP) {
                            if (name == null) {
                                name = tk.sval;
                                checkMethodOrFieldSpec(name);
                            } else {
                                String obfName = tk.sval;
                                checkJavaIdentifier(obfName);
                                entry = new RgsEntry(directive, name);
                                entry.obfName = obfName;
                            }
                        }
                    } else if (directive == RgsEntry.TYPE_CLASS) {
                        if (tk.sval.equals(OPTION_PUBLIC)) {
                            if (entry.retainToPublic || entry.retainToProtected) {
                                throw new Exception();
                            }
                            entry.retainToPublic = true;
                        } else if (tk.sval.equals(OPTION_PROTECTED)) {
                            if (entry.retainToPublic || entry.retainToProtected) {
                                throw new Exception();
                            }
                            entry.retainToProtected = true;
                        } else if (tk.sval.equals(OPTION_FIELD)) {
                            if ((!entry.retainToPublic && !entry.retainToProtected) || entry.retainMethodsOnly) {
                                throw new Exception();
                            }
                            entry.retainFieldsOnly = true;
                        } else if (tk.sval.equals(OPTION_METHOD)) {
                            if ((!entry.retainToPublic && !entry.retainToProtected) || entry.retainFieldsOnly) {
                                throw new Exception();
                            }
                            entry.retainMethodsOnly = true;
                        } else {
                            throw new Exception();
                        }
                    } else {
                        throw new Exception();
                    }
                } else if (ttype == StreamTokenizer.TT_EOL) {
                    if (entry != null) {
                        break;
                    }
                } else {
                    throw new Exception();
                }
            }
            next = entry;
        } catch (Exception e) {
            // Discard to end of erroneous line
            try {
                while ((ttype = tk.nextToken()) != StreamTokenizer.TT_EOF) {
                    if (ttype == StreamTokenizer.TT_EOL) {
                        break;
                    }
                }
            } catch (IOException ee) {
                // Take no action if StreamTokenizer fails here
            }

            // DEBUG
            //e.printStackTrace();

            // Save exception for throw from nextEntry()
            nextException = new Exception("Parser error at line " + Integer.toString(tk.lineno()) + " of script file.");
        }
    }

    // Throw if invalid
    private void checkMethodDescriptor(String s) throws Exception
    {
        if (s.length() == 0 || s.charAt(0) != '(')
        {
            throw new Exception();
        }
        s = s.substring(1);

        // Check each type
        while (s.length() > 0 && s.charAt(0) != ')')
        {
            s = checkFirstJavaType(s);
        }
        checkJavaType(s.substring(1));
    }

    // Throw if first type is invalid, else return all but first type in String
    private String checkFirstJavaType(String s) throws Exception
    {
        // Pull off the array specifiers
        while (s.charAt(0) == '[')
        {
            s = s.substring(1);
            if (s.length() == 0)
            {
                throw new Exception();
            }
        }

        // Check a type
        int pos = 0;
        switch (s.charAt(0))
        {
        case 'B':
        case 'C':
        case 'D':
        case 'F':
        case 'I':
        case 'J':
        case 'S':
        case 'V':
        case 'Z':
            break;

        case 'L':
            pos = s.indexOf(';');
            if (pos == -1)
            {
                throw new Exception();
            }
            // Check the class type
            checkClassSpec(s.substring(0, pos));
            break;

        default:
            throw new Exception();
        }
        return s.substring(pos + 1);
    }

    // Throw if type is invalid
    private void checkJavaType(String s) throws Exception
    {
        if (!checkFirstJavaType(s).equals(""))
        {
            throw new Exception();
        }
    }

    // Throw if invalid
    private void checkMethodOrFieldSpec(String s) throws Exception
    {
        if (s.length() == 0)
        {
            throw new Exception();
        }

        // Check the method or field name
        int pos = s.lastIndexOf('/');
        if (pos == -1)
        {
            throw new Exception();
        }
        checkJavaIdentifier(s.substring(pos + 1));
        checkClassSpec(s.substring(0, pos));
    }

    // Throw if invalid
    private void checkClassSpec(String s) throws Exception
    {
        if (s.length() == 0)
        {
            throw new Exception();
        }

        int pos = -1;
        while ((pos = s.lastIndexOf('$')) != -1)
        {
            checkJavaInnerIdentifier(s.substring(pos + 1));
            s = s.substring(0, pos);
        }
        while ((pos = s.lastIndexOf('/')) != -1)
        {
            checkJavaIdentifier(s.substring(pos + 1));
            s = s.substring(0, pos);
        }
        checkJavaIdentifier(s);
    }

    // Throw if invalid
    private void checkClassWCSpec(String s) throws Exception
    {
        // Check for wildcard package spec first
        if (s.length() == 0)
        {
            throw new Exception();
        }
        if (s.charAt(s.length() - 1) == '*')
        {
            if (!s.equals("*"))
            {
                if (s.length() < 3 || s.charAt(s.length() - 2) != '/')
                {
                    throw new Exception();
                }
                else
                {
                    s = s.substring(0, s.length() - 2);
                    int pos = -1;
                    while ((pos = s.lastIndexOf('/')) != -1)
                    {
                        checkJavaIdentifier(s.substring(pos + 1));
                        s = s.substring(0, pos);
                    }
                    checkJavaIdentifier(s);
                }
            }
        }
        else
        {
            // Check for regular class spec
            checkClassSpec(s);
        }
    }

    // Throw if invalid
    private void checkJavaIdentifier(String s) throws Exception
    {
        if (s.length() == 0 || !Character.isJavaIdentifierStart(s.charAt(0)))
        {
            throw new Exception();
        }
        for (int i = 1; i < s.length(); i++)
        {
            if (!Character.isJavaIdentifierPart(s.charAt(i)))
            {
                throw new Exception();
            }
        }
    }

    // Throw if invalid (allows for anon. inner class names like '4')
    private void checkJavaInnerIdentifier(String s) throws Exception
    {
        if (s.length() == 0)
        {
            throw new Exception();
        }
        for (int i = 0; i < s.length(); i++)
        {
            if (!Character.isJavaIdentifierPart(s.charAt(i)))
            {
                throw new Exception();
            }
        }
    }
}
