/* ===========================================================================
 * $Workfile: $
 * ===========================================================================
 *
 * RetroGuard -- an obfuscation package for Java classfiles.
 *
 * Copyright (c) 1999 Mark Welsh (markw@retrologic.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * The author may be contacted at markw@retrologic.com 
 *
 *
 * $Modtime: $
 * $Revision: 1.2 $
 */
package COM.rl.obf.patch;

import java.io.*;
import java.util.*;
import COM.rl.util.*;
import COM.rl.obf.*;

/**
 * Database of package and class relations formed from an RGS log file.
 *
 * @author Mark Welsh
 */
public class ClassDB {
    // Constants -------------------------------------------------------------


    // Fields ----------------------------------------------------------------
    private Pk root = null;   // Root package (Java default package)


    // Class Methods ---------------------------------------------------------


    // Instance Methods ------------------------------------------------------
    /** Ctor. */
    public ClassDB(File rgsFile) throws Exception {
        try {
            if (rgsFile != null && rgsFile.exists()) {
                InputStream rgsInputStream = new FileInputStream(rgsFile);
                createHierarchy(new RgsEnum(rgsInputStream));
                rgsInputStream.close();
            }
        } catch (Exception e) {
            // Reset the database
            root = null;

            // DEBUG
            //e.printStackTrace();

            throw new Exception("Log-file corrupt or in an older format (pre-v1.1)");
        }
    }

    /** Convert entries in file to obfuscated versions. */
    public Vector toObf(File listFile) throws Exception {
        Vector outNames = new Vector();
        
        // Run through the listed names
        for (Enumeration enum = getInNames(listFile); 
             enum.hasMoreElements(); ) {
            // Convert the obfuscated portions
            String inName = (String)enum.nextElement();
            String outName = null;
            
            // Ignore inner class listings
            if (inName.indexOf('$') == -1) {
                // Split at first '.'
                int pos = inName.indexOf('.');
                String pre = (pos == -1 ? inName : inName.substring(0, pos));
                String post = (pos == -1 ? null : inName.substring(pos + 1));
                if (post != null && post.equals("class")) {
                    // Convert class name
                    outName = getOutName(pre) + ".class";
                } else {
                    // Convert resource name
                    outName = getOutName(inName);
                }
                
                // Store the converted name
                if (outName != null) {
                    outNames.addElement(outName);
                }
            }
        }
        return outNames;
    }

    // Map a name
    private String getOutName(String inName) {
        try {
            TreeItem ti = root;
            StringBuffer sb = new StringBuffer();
            for (Enumeration nameEnum = ClassTree.getNameEnum(inName); 
                 nameEnum.hasMoreElements(); ) {
                Cons nameSegment = (Cons)nameEnum.nextElement();
                char tag = ((Character)nameSegment.car).charValue();
                String name = (String)nameSegment.cdr;
                switch (tag) {
                case ClassTree.PACKAGE_LEVEL:
                    if (ti != null) {
                        ti = ((Pk)ti).getPackage(name);
                        if (ti != null) {
                            sb.append(ti.getOutName());
                        } else {
                            sb.append(name); 
                        }
                    } else {
                        sb.append(name);
                    }
                    sb.append(ClassTree.PACKAGE_LEVEL);
                    break;
                    
                case ClassTree.CLASS_LEVEL:
                    if (ti != null) {
                        ti = ((Pk)ti).getClass(name);
                        if (ti != null) {
                            sb.append(ti.getOutName());
                        } else {
                            sb.append(name); 
                        }
                    } else {
                        sb.append(name);
                    }
                    return sb.toString();
                    
                default:
                    throw new Exception("Internal error: illegal package/class name tag");
                }
            }
        } catch (Exception e) {
            // Just drop through and return the original name
        }
        return inName;
    }

    /** Convert entries in file to obfuscated versions. */
    public Vector toObfForConv(File listFile) throws Exception {
        Vector outNames = new Vector();
        
        // Run through the listed names
        for (Enumeration enum = getInNames(listFile); 
             enum.hasMoreElements(); ) {
            // Convert the obfuscated portions
            String inName = (String)enum.nextElement();
            String outName = null;
            
            // Split at first '.'
            int pos = inName.indexOf('.');
            String pre = (pos == -1 ? inName : inName.substring(0, pos));
            String post = (pos == -1 ? null : inName.substring(pos + 1));
            if (post != null && post.equals("class")) {
                // Convert class name
                outName = getOutNameForConv(pre) + ".class";
            } else {
                // Convert resource name
                outName = getOutNameForConv(inName);
            }
            
            // Store the converted name
            if (outName != null) {
                outNames.addElement(outName);
            }
        }
        return outNames;
    }

    // Map a name
    private String getOutNameForConv(String inName) {
        try {
            TreeItem ti = root;
            StringBuffer sb = new StringBuffer();
            for (Enumeration nameEnum = ClassTree.getNameEnum(inName); 
                 nameEnum.hasMoreElements(); ) {
                Cons nameSegment = (Cons)nameEnum.nextElement();
                char tag = ((Character)nameSegment.car).charValue();
                String name = (String)nameSegment.cdr;
                switch (tag) {
                case ClassTree.PACKAGE_LEVEL:
                    if (ti != null) {
                        ti = ((Pk)ti).getPackage(name);
                        if (ti != null) {
                            sb.append(ti.getOutName());
                        } else {
                            sb.append(name); 
                        }
                    } else {
                        sb.append(name);
                    }
                    sb.append(ClassTree.PACKAGE_LEVEL);
                    break;
                    
                case ClassTree.CLASS_LEVEL:
                    if (ti != null) {
                        ti = ((PkCl)ti).getClass(name);
                        if (ti != null) {
                            sb.append(ti.getOutName());
                        } else {
                            sb.append(name); 
                        }
                    } else {
                        sb.append(name);
                    }
                    if (nameEnum.hasMoreElements()) {
                        sb.append(ClassTree.CLASS_LEVEL);
                    } else {
                        return sb.toString();
                    }
                    break;
                    
                default:
                    throw new Exception("Internal error: illegal package/class name tag");
                }
            }
        } catch (Exception e) {
            // Just drop through and return the original name
        }
        return inName;
    }

    // Parse the file names from the list
    private Enumeration getInNames(File listFile) throws Exception {
        InputStream is = listFile == null ? System.in : 
            new FileInputStream(listFile);
        StreamTokenizer tk = new StreamTokenizer(
                             new BufferedReader(
                             new InputStreamReader(is)));
        tk.resetSyntax();
        tk.whitespaceChars(0x00, 0x20);
        tk.wordChars('*', '*');
        tk.wordChars('.', '.');
        tk.wordChars(';', ';');
        tk.wordChars('_', '_');
        tk.wordChars('[', '[');
        tk.wordChars('(', ')');
        tk.wordChars('$', '$');
        tk.wordChars('/', '9');
        tk.wordChars('A', 'Z');
        tk.wordChars('a', 'z');
        tk.commentChar('#');
        tk.eolIsSignificant(false);
        Vector inNames = new Vector();
        try {
            int ttype;
            while ((ttype = tk.nextToken()) != StreamTokenizer.TT_EOF) {
                if (ttype == StreamTokenizer.TT_WORD) {
                    inNames.addElement(tk.sval);
                }
            }
        } finally {
            return inNames.elements();
        }
    }

    // Create database from RGS log entries.
    private void createHierarchy(RgsEnum rgsEnum) throws Exception {
        // Create the root of the package hierarchy
        root = new Pk(null, "");

        // Enumerate the entries in the RGS script
        while (rgsEnum.hasMoreEntries()) {
            RgsEntry entry = rgsEnum.nextEntry();
            switch (entry.type) {
            case RgsEntry.TYPE_PACKAGE_MAP:
                addPackage(entry.name, entry.obfName);
                break;
                
            case RgsEntry.TYPE_CLASS:
                addClass(entry.name, getLastIdentifier(entry.name));
                break;
                
            case RgsEntry.TYPE_CLASS_MAP:
                addClass(entry.name, entry.obfName);
                break;
                
            case RgsEntry.TYPE_ATTR:
            case RgsEntry.TYPE_METHOD:
            case RgsEntry.TYPE_FIELD:
            case RgsEntry.TYPE_METHOD_MAP:
            case RgsEntry.TYPE_FIELD_MAP:
                // Ignore attribute, method and field entries
                break;
                
            default:
                throw new Exception("Illegal type received from the .rgs script");
            }
        }
    }

    // Add a package to the hierarchy
    private void addPackage(String fullName, String obfName) throws Exception {
        TreeItem ti = root;
        for (Enumeration nameEnum = ClassTree.getNameEnum(fullName); 
             nameEnum.hasMoreElements(); ) {
            Cons nameSegment = (Cons)nameEnum.nextElement();
            String name = (String)nameSegment.cdr;
            ti = ((Pk)ti).addPackage(name);
        }

        // Set the obfuscated name for the package
        ti.setOutName(obfName);
    }

    // Add a class to the hierarchy
    private void addClass(String fullName, String obfName) throws Exception {
        TreeItem ti = root;
        for (Enumeration nameEnum = ClassTree.getNameEnum(fullName); 
             nameEnum.hasMoreElements(); ) {
            Cons nameSegment = (Cons)nameEnum.nextElement();
            char tag = ((Character)nameSegment.car).charValue();
            String name = (String)nameSegment.cdr;
            switch (tag) {
            case ClassTree.PACKAGE_LEVEL:
                ti = ((Pk)ti).addPackage(name);
                break;

            case ClassTree.CLASS_LEVEL:
                // If inner class, just add placeholder classes up the tree
                if (nameEnum.hasMoreElements()) {
                    ti = ((PkCl)ti).addPlaceholderClass(name);
                } else {
                    ti = ((PkCl)ti).addClass(name, null, null);
                }
                break;

            default:
                throw new Exception("Internal error: illegal package/class name tag");
            }
        }

        // Set the obfuscated name for the class (anonymous inner classes 
        // are already set, so leave them be)
        if (obfName != null && !Character.isDigit(obfName.charAt(0))) {
            ti.setOutName(obfName);
        }
    }

    // Return the last Java identifier from the passed String
    private String getLastIdentifier(String name) {
        String result = null;
        if (name != null) {
            int outer = name.lastIndexOf('/');
            int inner = name.lastIndexOf('$');
            if (outer != -1 || inner != -1) {
                int pos = Math.max(inner, outer);
                result = name.substring(pos + 1);
            }
        }
        return result;
    }
}

