/* ===========================================================================
 * $RCSfile: RGconv.java,v $
 * ===========================================================================
 *
 * RetroGuard -- an obfuscation package for Java classfiles.
 *
 * Copyright (c) 1999 Mark Welsh (markw@retrologic.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * The author may be contacted at markw@retrologic.com 
 *
 *
 * $Date: 1998/10/01 14:14:12 $
 * $Revision: 1.2 $
 */

import java.io.*;
import java.util.*;
import COM.rl.obf.*;
import COM.rl.obf.patch.*;

/**
 * Main class for conversion utility. Converts from normal to obfuscated
 * class filenames. 
 *
 * @author      Mark Welsh
 */
public class RGconv {
    // Constants -------------------------------------------------------------


    // Fields ----------------------------------------------------------------
    private File rgsFile;
    private File inListFile;
    private File outListFile;


    // Class Methods ---------------------------------------------------------
    /**
     * Main entry point for RGconv - runs with obligatory 3 arguments. 
     */
    public static void main(String args[]) {
        // Check arg-list for validity
        try {
            // Get the arguments, or set their default values
            if (args.length < 1 || args.length > 3) {
                throw new IllegalArgumentException("Invalid number of arguments.");
            }
            File rgsFile = new File(args[0]);
            File inListFile = args.length > 1 ? new File(args[1]) : null;
            File outListFile = args.length > 2 ? new File(args[2]) : null;

            // Script/log file must exist and be readable
            if (!rgsFile.exists()) {
                throw new IllegalArgumentException("Log file specified does not exist.");
            }
            if (!rgsFile.canRead()) {
                throw new IllegalArgumentException("Log file specified exists but cannot be read.");
            }

            // Input list file must exist and be readable
            if (inListFile != null && !inListFile.exists()) {
                throw new IllegalArgumentException("Input list specified does not exist.");
            }
            if (inListFile != null && !inListFile.canRead()) {
                throw new IllegalArgumentException("Input list specified exists but cannot be read.");
            }

            // Output list file must be writable if it exists
            if (outListFile != null && outListFile.exists() && !outListFile.canWrite()) {
                throw new IllegalArgumentException("Output list file cannot be written to.");
            }

            // Call the main entry point on the patch generator.
            RGconv.convert(rgsFile, inListFile, outListFile);
        } catch (IllegalArgumentException e) {
            System.err.println("Problem: " + (e.getMessage() != null ? e.getMessage() : ""));
            System.err.println("Usage: java RGconv RGS-FILE INPUT-FILE OUTPUT-FILE");
            System.err.println("  where RGS-FILE is the log file from the obfuscation");
            System.err.println("        INPUT-FILE is a list of unobfuscated class filenames");
            System.err.println("        OUTPUT-FILE is name for the list of obfuscated class filenames");
            System.exit(-1);
        } catch (Exception e) {
            System.err.println("RGconv error: " + e.toString());
            System.exit(-1);
        }
        System.exit(0);
    }

    /**
     * Main entry point for the converter.
     *
     * @param rgsFile log-file from the obfuscation.
     * @param inListFile list of unobfuscated class files
     * @param outListFile file for list of obfuscated class files
     */
    public static void convert(File rgsFile, File inListFile, 
                               File outListFile) throws Exception {
        // Create the name mapping database, using the log file
        ClassDB db = new ClassDB(rgsFile);

        // Convert the list file to obfuscated form
        Vector output = db.toObfForConv(inListFile);

        // Write the output file
        PrintStream out = null;
        try {
            out = outListFile == null ? System.out :
                new PrintStream(
                new BufferedOutputStream(
                new FileOutputStream(outListFile)));
            for (Enumeration enum = output.elements(); 
                 enum.hasMoreElements(); ) {
                out.println((String)enum.nextElement());
            }
            out.flush();
        } finally {
            if (out != null) {
                out.close();
            }
        }
    }
}
