/* ===========================================================================
 * $RCSfile: RGpatch.java,v $
 * ===========================================================================
 *
 * RetroGuard -- an obfuscation package for Java classfiles.
 *
 * Copyright (c) 1999 Mark Welsh (markw@retrologic.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * The author may be contacted at markw@retrologic.com 
 *
 *
 * $Date: 1998/10/01 14:14:12 $
 * $Revision: 1.2 $
 */

import java.io.*;
import java.util.*;
import COM.rl.obf.*;
import COM.rl.obf.patch.*;

/**
 * Main class for patch utility. Creates a patch Jar from a full 
 * obfuscation run. 
 *
 * @author      Mark Welsh
 */
public class RGpatch {
    // Constants -------------------------------------------------------------


    // Fields ----------------------------------------------------------------
    private File inFile;
    private File outFile;
    private File rgsFile;
    private File listFile;


    // Class Methods ---------------------------------------------------------
    /**
     * Main entry point for RGpatch - runs with obligatory 4 arguments. 
     */
    public static void main(String args[]) {
        // Check arg-list for validity
        try {
            // Get the arguments, or set their default values
            if (args.length != 4) {
                throw new IllegalArgumentException("Invalid number of arguments.");
            }
            File inFile = new File(args[0]);
            File outFile = new File(args[1]);
            File rgsFile = new File(args[2]);
            File listFile = new File(args[3]);

            // Input JAR file must exist and be readable
            if (!inFile.exists()) {
                throw new IllegalArgumentException("JAR specified for obfuscation does not exist.");
            }
            if (!inFile.canRead()) {
                throw new IllegalArgumentException("JAR specified for obfuscation exists but cannot be read.");
            }

            // Output JAR file must be writable if it exists
            if (outFile.exists() && !outFile.canWrite()) {
                throw new IllegalArgumentException("Output JAR file cannot be written to.");
            }

            // Script/log file must exist and be readable
            if (!rgsFile.exists()) {
                throw new IllegalArgumentException("Log file specified does not exist.");
            }
            if (!rgsFile.canRead()) {
                throw new IllegalArgumentException("Log file specified exists but cannot be read.");
            }

            // List file must exist and be readable
            if (!listFile.exists()) {
                throw new IllegalArgumentException("List specified does not exist.");
            }
            if (!listFile.canRead()) {
                throw new IllegalArgumentException("List specified exists but cannot be read.");
            }

            // Call the main entry point on the patch generator.
            RGpatch.patch(inFile, outFile, rgsFile, listFile);
        } catch (IllegalArgumentException e) {
            System.err.println("Problem: " + (e.getMessage() != null ? e.getMessage() : ""));
            System.err.println("Usage: java RGpatch INPUT-FILE OUTPUT-FILE LOG-FILE LIST-FILE");
            System.err.println("  where INPUT-FILE is the full obfuscated JAR");
            System.err.println("        OUTPUT-FILE is name for the patch JAR");
            System.err.println("        RGS-FILE is the log file from the obfuscation");
            System.err.println("        LIST-FILE is the list of files (unobfuscated names) to include in the patch");
            System.exit(-1);
        } catch (Exception e) {
            System.err.println("RGpatch error: " + e.getMessage());
            System.exit(-1);
        }
        System.exit(0);
    }

    /**
     * Main entry point for the patch generator.
     *
     * @param inFile the full, obfuscated JAR
     * @param outFile a writable file for obfuscated patch JAR
     * @param rgsFile log-file from the obfuscation.
     * @param listFile list of files to include in the patch (unobfuscated names)
     */
    public static void patch(File inFile, File outFile, 
                             File rgsFile, File listFile) throws Exception {
        // Create the name mapping database, using the log file
        ClassDB db = new ClassDB(rgsFile);

        // Convert the list file to obfuscated form, and initialize the 
        // patch generator
        PatchMaker pm = new PatchMaker(db.toObf(listFile));

        // Copy the obfuscated Jar to a patch Jar, adding only classes (and
        // all of their inner classes) known to the PatchMaker
        pm.makePatch(inFile, outFile);
    }
}
