/* ===========================================================================
 * $RCSfile: RetroGuard.java,v $
 * ===========================================================================
 *
 * RetroGuard -- an obfuscation package for Java classfiles.
 *
 * Copyright (c) 1999 Mark Welsh (markw@retrologic.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * The author may be contacted at markw@retrologic.com 
 *
 *
 * $Date: 1998/10/01 14:14:12 $
 * $Revision: 1.2 $
 */

import java.io.*;
import java.util.*;
import COM.rl.obf.*;
import COM.rl.obf.gui.*;

/**
 * Main class for obfuscator package. All methods exposed through the JAR belong
 * to this class.
 *
 * @author      Mark Welsh
 */
public class RetroGuard
{
    // Constants -------------------------------------------------------------
    public static final String DEFAULT_IN_FILE_NAME = "in.jar";
    public static final String DEFAULT_OUT_FILE_NAME = "out.jar";
    public static final String DEFAULT_RGS_FILE_NAME = "script.rgs";
    public static final String DEFAULT_LOG_FILE_NAME = "retroguard.log";
    private static final String LOG_TITLE_PRE_VERSION = "# RetroGuard Bytecode Obfuscator, v";
    private static final String LOG_TITLE_POST_VERSION = ", a product of Retrologic Systems - www.retrologic.com";
    private static final String LOG_CREATED = "# Logfile created on ";
    private static final String LOG_INPUT_FILE =  "# Jar file to be obfuscated:           ";
    private static final String LOG_OUTPUT_FILE = "# Target Jar file for obfuscated code: ";
    private static final String LOG_SCRIPT_FILE = "# RetroGuard Script file used:         ";
    private static final String LOG_NO_SCRIPT = "(none, defaults used)";
    private static final String LOG_ERROR = "# Unrecoverable error during obfuscation:";
    private static final String SEE_LOG_FILE = "Unrecoverable error during obfuscation, see log file for details.";


    // Fields ----------------------------------------------------------------
    private File inFile;
    private File outFile;
    private File rgsFile;
    private File logFile;


    // Class Methods ---------------------------------------------------------
    /**
     * Obfuscate an input Jar file on the file system into an output Jar file,
     * using the script file supplied to change obfuscation settings from their
     * default. <p>
     *
     * Default setting are: remove all attributes except 'Code', 'ConstantValue',
     * 'Exceptions', 'InnerClasses', 'Synthetic'; obfuscate all package, class,
     * interface, method, field names in the Jar except where this would break
     * polymorphism from classes, interfaces outside of the Jar. <p>
     *
     * Usage: java RetroGuard [INPUT-FILE [OUTPUT-FILE [SCRIPT-FILE [LOG-FILE]]]]
     * where INPUT-FILE is the name of the JAR to be obfuscated (defaults to 'in.jar'),
     * OUTPUT-FILE is the name for the obfuscated JAR (defaults to 'out.jar'),
     * SCRIPT-FILE is the file name of a valid RetroGuard script (*.rgs) file
     * (defaults to 'script.rgs'), LOG-FILE is the file name of the output log file
     * in which name mappings are listed (defaults to 'retroguard.log'). <p>
     *
     * BNF grammar for the script file is: <br><br>
     *
     * \<script\> ::= \<script_line\>* <br>
     * \<script_line\> ::= [\<comment\> | \<exclusion\>] "\n" <br>
     * \<exclusion\> ::= (\<attribute_exc\> | \<class_exc\> | \<method_exc\> | \<field_exc\>) [\<comment\>] <br>
     * \<comment\> ::= "#" \<character\>* <br>
     * \<character\> ::= "\040"-"\176" <br>
     * \<attribute_exc\> ::= ".attribute" ("SourceFile" | "LocalVariableTable" | "LineNumberTable") <br>
     * \<class_exc\> ::= ".class" \<class_spec\> [("public" | "protected") ["method" | "field"]] <br>
     * \<method_exc\> ::= ".method" \<method_field_spec\> \<method_descriptor\> <br>
     * \<field_exc\> ::= ".field" \<method_field_spec\> \<java_type\> <br>
     * \<method_field_spec\> ::= \<class_spec\>"/"\<java_identifier\> <br>
     * \<method_descriptor\> ::= "("\<java_type\>*")"\<java_type\> <br>
     * \<java_type\> ::= ("["\<java_type\>) | "B" | "C" | "D" | "F" | "I" | "J" | "S" | "V" | "Z" | ("L"\<class_spec\>";") <br>
     * \<class_spec\> ::= ((\<java_identifier\>"/")*)((\<java_identifier\>"$")*)\<java_identifier\> <br>
     * \<java_identifier\> ::= \<java_letter\>\<java_letter_or_digit\>* // but not a Java keyword, or 'true' or 'false' <br>
     * // \<java_letter\> is a Unicode character for which java.lang.Character.isJavaIdentifierStart() is true <br>
     * // \<java_letter_or_digit\> is a Unicode character for which java.lang.Character.isJavaIdentifierPart() is true <br>
     *
     */
    public static void main(String args[])
    {
        // Check arg-list for validity
        try
        {
            // Get the arguments, or set their default values
            if (args.length < 0 || args.length > 4)
            {
                throw new IllegalArgumentException("Invalid number of arguments.");
            }
            File inFile = new File(args.length < 1 ? DEFAULT_IN_FILE_NAME : args[0]);
            File outFile = new File(args.length < 2 ? DEFAULT_OUT_FILE_NAME : args[1]);
            File rgsFile = new File(args.length < 3 ? DEFAULT_RGS_FILE_NAME : args[2]);
            File logFile = new File(args.length < 4 ? DEFAULT_LOG_FILE_NAME : args[3]);

            // Input JAR file must exist and be readable
            if (!inFile.exists())
            {
                throw new IllegalArgumentException("JAR specified for obfuscation does not exist.");
            }
            if (!inFile.canRead())
            {
                throw new IllegalArgumentException("JAR specified for obfuscation exists but cannot be read.");
            }

            // Output JAR file must be writable if it exists
            if (outFile.exists() && !outFile.canWrite())
            {
                throw new IllegalArgumentException("Output JAR file cannot be written to.");
            }

            // Script file must be readable if it exists, but need not exist (revert to default
            // obfuscation settings in that case)
            if (rgsFile.exists())
            {
                if (!rgsFile.canRead())
                {
                    throw new IllegalArgumentException("Script file exists but cannot be read.");
                }
            }

            // Logfile must be writable if it exists
            if (logFile.exists() && !logFile.canWrite())
            {
                throw new IllegalArgumentException("Logfile cannot be written to.");
            }

            // Call the main entry point on the obfuscator.
            RetroGuard.obfuscate(inFile, outFile, rgsFile, logFile);
            System.exit(0);
        }
        catch (IllegalArgumentException e)
        {
            System.err.println("Problem: " + (e.getMessage() != null ? e.getMessage() : ""));
            System.err.println("Usage: java RetroGuard [INPUT-FILE [OUTPUT-FILE [SCRIPT-FILE [LOG-FILE]]]]");
            System.err.println("  where INPUT-FILE is the JAR to be obfuscated (default: 'in.jar')");
            System.err.println("        OUTPUT-FILE is name for the obfuscated JAR (default: 'out.jar')");
            System.err.println("        SCRIPT-FILE is a valid RetroGuard script (default: 'script.rgs').");
            System.err.println("        LOG-FILE is the name for the log file (default: 'retroguard.log').");
            System.err.println("Usage: 'java RGgui' starts the wizard for constructing script files.");
            System.exit(-1);
        }
        catch (Exception e)
        {
            System.err.println("RetroGuard error: " + e.toString());
            System.exit(-1);
        }
    }

    /**
     * Main entry point for the obfuscator.
     *
     * @param inFile a File pointing to a readable JAR
     * @param outFile a writable JAR for obfuscated output
     * @param rgsFile valid RetroGuard Script data, or null (which implies default settings)
     * @param logFile PrintWriter for the log data.
     */
    public static void obfuscate(File inFile, File outFile, File rgsFile, File logFile) throws Exception
    {
        new RetroGuard(inFile, outFile, rgsFile, logFile).run();
    }


    // Instance Methods ------------------------------------------------------
    // Private constructor takes in-jar, out-jar and script specifiers.
    private RetroGuard(File inFile, File outFile, File rgsFile, File logFile)
    {
        this.inFile = inFile;
        this.outFile = outFile;
        this.rgsFile = rgsFile;
        this.logFile = logFile;
    }

    // Run the obfuscator.
    private void run() throws Exception
    {
        // Create the session log file
        PrintWriter log = new PrintWriter(
                          new BufferedOutputStream(
                          new FileOutputStream(logFile)));
        try
        {
            // Write out the log header
            writeLogHeader(log);

            // Create the name mapping database for the input JAR, constrained by the
            // options in the rgs script
            GuardDB db = new GuardDB(inFile);
            try
            {
                if (rgsFile.exists())
                {
                    InputStream rgsInputStream = new FileInputStream(rgsFile);
                    db.retain(new RgsEnum(rgsInputStream), log);
                    rgsInputStream.close();
                }
                db.remapTo(outFile, log);
            }
            finally
            {
                db.close();
            }
        }
        catch (Exception e)
        {
            // Log exceptions before exiting
            log.println();
            log.println(LOG_ERROR);
            log.println("# " + e.toString());
            log.println();
            System.err.println(SEE_LOG_FILE);
            /* DEBUG e.printStackTrace(System.err); */
        }
        finally
        {
            log.flush();
            log.close();
        }
    }

    // Write a header out to the log file
    private void writeLogHeader(PrintWriter log)
    {
        log.println("# If this log is to be used for incremental obfuscation / patch generation, ");
        log.println("# add any '.class', '.method', '.field' and '.attribute' restrictions here:");
        log.println();
        log.println();
        log.println("#-DO-NOT-EDIT-BELOW-THIS-LINE------------------DO-NOT-EDIT-BELOW-THIS-LINE--");
        log.println("#");
        log.println(LOG_TITLE_PRE_VERSION + Version.getVersion() + LOG_TITLE_POST_VERSION);
        log.println("#");
        log.println(LOG_CREATED + new Date().toString());
        log.println("#");
        log.println(LOG_INPUT_FILE + inFile.getName());
        log.println(LOG_OUTPUT_FILE + outFile.getName());
        log.println(LOG_SCRIPT_FILE + (rgsFile.exists() ? rgsFile.getName() : LOG_NO_SCRIPT));
        log.println("#");
    }
}
